"""Scryfall object models."""

from dataclasses import dataclass
import datetime as dt
from decimal import Decimal
from enum import Enum
from typing import ClassVar
from typing import Dict
from typing import NewType
from typing import Optional
from typing import Sequence
from uuid import UUID

URI = NewType("URI", str)


class ScryObject:
    """Base object class for scryfall response objects."""

    object: ClassVar[str] = "object"


class ScryColor(str, Enum):
    """Enum for https://scryfall.com/docs/api/colors#color-arrays"""

    WHITE = "W"
    BLUE = "U"
    BLACK = "B"
    RED = "R"
    GREEN = "G"


class ScrySetType(str, Enum):
    """Enum for https://scryfall.com/docs/api/sets#set-types"""

    CORE = "core"
    EXPANSION = "expansion"
    MASTERS = "masters"
    MASTERPIECE = "masterpiece"
    FROM_THE_VAULT = "from_the_vault"
    SPELLBOOK = "spellbook"
    PREMIUM_DECK = "premium_deck"
    DUEL_DECK = "duel_deck"
    DRAFT_INNOVATION = "draft_innovation"
    TREASURE_CHEST = "treasure_chest"
    COMMANDER = "commander"
    PLANECHASE = "planechase"
    ARCHENEMY = "archenemy"
    VANGUARD = "vanguard"
    FUNNY = "funny"
    STARTER = "starter"
    BOX = "box"
    PROMO = "promo"
    TOKEN = "token"
    MEMORABILIA = "memorabilia"


class ScryCardLayout(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#layout"""

    NORMAL = "normal"
    SPLIT = "split"
    FLIP = "flip"
    TRANSFORM = "transform"
    MELD = "meld"
    LEVELER = "leveler"
    SAGA = "saga"
    ADVENTURE = "adventure"
    PLANAR = "planar"
    SCHEME = "scheme"
    VANGUARD = "vanguard"
    TOKEN = "token"
    DOUBLE_FACED_TOKEN = "double_faced_token"
    EMBLEM = "emblem"
    AUGMENT = "augment"
    HOST = "host"
    ART_SERIES = "art_series"
    DOUBLE_SIDED = "double_sided"


class ScryCardFrame(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#frames"""

    Y1993 = "1993"
    Y1997 = "1997"
    Y2003 = "2003"
    Y2015 = "2015"
    FUTURE = "future"


class ScryFrameEffect(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#frame-effects"""

    NONE = ""
    LEGENDARY = "legendary"
    MIRACLE = "miracle"
    NYXTOUCHED = "nyxtouched"
    DRAFT = "draft"
    DEVOID = "devoid"
    TOMBSTONE = "tombstone"
    COLORSHIFTED = "colorshifted"
    SUNMOONDFC = "sunmoondfc"
    COMPASSLANDDFC = "compasslanddfc"
    ORIGINPWDFC = "originpwdfc"
    MOONELDRAZIDFC = "mooneldrazidfc"
    EXTENDEDART = "extendedart"
    INVERTED = "inverted"
    NYXBORN = "nyxborn"
    SHOWCASE = "showcase"
    WAXINGANDWANINGMOONDFC = "waxingandwaningmoondfc"


class ScryBorderColor(str, Enum):
    """Enum for card border_color"""

    BLACK = "black"
    BORDERLESS = "borderless"
    GOLD = "gold"
    SILVER = "silver"
    WHITE = "white"


class ScryGame(str, Enum):
    """Enum for card games"""

    PAPER = "paper"
    ARENA = "arena"
    MTGO = "mtgo"


class ScryRarity(str, Enum):
    """Enum for card rarity"""

    COMMON = "common"
    UNCOMMON = "uncommon"
    RARE = "rare"
    MYTHIC = "mythic"


class ScryFormat(str, Enum):
    """Enum for card legalities keys"""

    BRAWL = "brawl"
    COMMANDER = "commander"
    DUEL = "duel"
    FRONTIER = "frontier"
    FUTURE = "future"
    LEGACY = "legacy"
    MODERN = "modern"
    OLDSCHOOL = "oldschool"
    PAUPER = "pauper"
    PENNY = "penny"
    STANDARD = "standard"
    VINTAGE = "vintage"
    HISTORIC = "historic"
    PIONEER = "pioneer"


class ScryLegality(str, Enum):
    """Enum for card legalities values"""

    LEGAL = "legal"
    NOT_LEGAL = "not_legal"
    RESTRICTED = "restricted"
    BANNED = "banned"


@dataclass(frozen=True)
class ScryObjectList(ScryObject):
    """Model for https://scryfall.com/docs/api/lists"""

    object: ClassVar[str] = "list"
    data: Sequence[ScryObject]
    has_more: bool
    next_page: Optional[URI]
    total_cards: Optional[int]
    warnings: Optional[Sequence[str]]


@dataclass(frozen=True)
class ScrySet(ScryObject):
    """Model for https://scryfall.com/docs/api/sets"""

    object: ClassVar[str] = "set"
    id: UUID
    code: str
    mtgo_code: Optional[str]
    arena_code: Optional[str]
    tcgplayer_id: Optional[int]
    name: str
    set_type: ScrySetType
    released_at: Optional[dt.date]
    block_code: Optional[str]
    block: Optional[str]
    parent_set_code: Optional[str]
    card_count: int
    digital: bool
    foil_only: bool
    icon_svg_uri: URI
    search_uri: URI
    scryfall_uri: URI
    uri: URI


@dataclass(frozen=True)
class ScryRelatedCard(ScryObject):
    """Model for https://scryfall.com/docs/api/cards#related-card-objects"""

    object: ClassVar[str] = "related_card"
    id: UUID
    component: str
    name: str
    type_line: str
    uri: URI


@dataclass(frozen=True)
class ScryCardFace(ScryObject):
    """Model for https://scryfall.com/docs/api/cards#card-face-objects"""

    object: ClassVar[str] = "card_face"
    artist: Optional[str]
    artist_id: Optional[UUID]
    color_indicator: Optional[Sequence[ScryColor]]
    colors: Optional[Sequence[ScryColor]]
    flavor_text: Optional[str]
    illustration_id: Optional[UUID]
    image_uris: Optional[Dict[str, URI]]
    loyalty: Optional[str]
    mana_cost: str
    name: str
    oracle_text: Optional[str]
    power: Optional[str]
    printed_name: Optional[str]
    printed_text: Optional[str]
    printed_type_line: Optional[str]
    toughness: Optional[str]
    type_line: Optional[str]
    watermark: Optional[str]


@dataclass(frozen=True)
class CardPreviewBlock:
    """Model for card preview block."""

    source: str
    source_uri: str  # TODO: should be uri
    previewed_at: dt.date


@dataclass(frozen=True)
class ScryCard(ScryObject):
    """Model for https://scryfall.com/docs/api/cards"""

    object: ClassVar[str] = "card"
    # Core Card Fields
    arena_id: Optional[int]
    id: UUID
    lang: str
    mtgo_id: Optional[int]
    mtgo_foil_id: Optional[int]
    multiverse_ids: Optional[Sequence[int]]
    tcgplayer_id: Optional[int]
    oracle_id: UUID
    prints_search_uri: URI
    rulings_uri: URI
    scryfall_uri: URI
    uri: URI
    # Gameplay Fields
    all_parts: Optional[Sequence[ScryRelatedCard]]
    card_faces: Optional[Sequence[ScryCardFace]]
    cmc: Decimal
    colors: Optional[Sequence[ScryColor]]
    color_identity: Sequence[ScryColor]
    color_indicator: Optional[Sequence[ScryColor]]
    edhrec_rank: Optional[int]
    foil: bool
    hand_modifier: Optional[str]
    layout: ScryCardLayout
    legalities: Dict[ScryFormat, ScryLegality]
    life_modifier: Optional[str]
    loyalty: Optional[str]
    mana_cost: Optional[str]
    name: str
    nonfoil: bool
    oracle_text: Optional[str]
    oversized: bool
    power: Optional[str]
    reserved: bool
    toughness: Optional[str]
    type_line: Optional[str]
    # Print Fields
    artist: Optional[str]
    artist_ids: Optional[Sequence[UUID]]
    booster: bool
    border_color: ScryBorderColor
    card_back_id: UUID
    collector_number: str
    digital: bool
    flavor_text: Optional[str]
    frame_effect: Optional[ScryFrameEffect]
    frame_effects: Optional[Sequence[ScryFrameEffect]]
    frame: ScryCardFrame
    full_art: bool
    games: Sequence[ScryGame]
    highres_image: bool
    illustration_id: Optional[UUID]
    image_uris: Optional[Dict[str, URI]]
    prices: Optional[Dict[str, Decimal]]  # TODO: enum keys
    printed_name: Optional[str]
    printed_text: Optional[str]
    printed_type_line: Optional[str]
    promo: bool
    promo_types: Optional[Sequence[str]]
    purchase_uris: Optional[Dict[str, URI]]
    rarity: ScryRarity
    related_uris: Optional[Dict[str, URI]]
    released_at: dt.date
    reprint: bool
    scryfall_set_uri: URI
    set_name: str
    set_search_uri: URI
    set_type: str
    set_uri: URI
    set: str
    story_spotlight: bool
    textless: bool
    variation: bool
    variation_of: Optional[UUID]
    watermark: Optional[str]
    preview: Optional[CardPreviewBlock]


@dataclass(frozen=True)
class ScryBulkData(ScryObject):
    """Model for https://scryfall.com/docs/api/bulk-data"""

    object: ClassVar[str] = "bulk_data"
    id: UUID
    type: str
    name: str
    description: str
    permalink_uri: URI
    updated_at: dt.datetime
    size: Optional[int]
    compressed_size: Optional[int]
    content_type: str
    content_encoding: str



"""Card and set index container."""

import collections
import string
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from uuid import UUID

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.mtg import util
from mtg_ssm.scryfall.models import ScryCard
from mtg_ssm.scryfall.models import ScrySet


def name_card_sort_key(card: ScryCard) -> Tuple[str, int, str]:
    """Key function for sorting cards in a by-name list."""
    card_num, card_var = util.collector_int_var(card)
    return (card.set, card_num or 0, card_var or "")  # TODO: sort by set release date


def set_card_sort_key(card: ScryCard) -> Tuple[int, str]:
    """Key function for sorting cards in a by-set list."""
    card_num, card_var = util.collector_int_var(card)
    return (card_num or 0, card_var or "")


def build_snnmas(
    card: ScryCard,
) -> Iterable[Tuple[Optional[str], str, Optional[str], Optional[int], Optional[str]]]:
    """Build set, name, number, multiverse id tuple keys."""
    names_cnums: Set[Tuple[str, Optional[str]]] = {(card.name, card.collector_number)}
    for i, card_face in enumerate(card.card_faces or ()):
        names_cnums |= {
            (card_face.name, card.collector_number),
            (card_face.name, card.collector_number + string.ascii_lowercase[i]),
        }
    names_cnums |= {(n, None) for n, _ in names_cnums}

    sets: Set[Optional[str]] = {card.set, None}

    mvids: Set[Optional[int]] = {None} | set(card.multiverse_ids or ())

    artists: Set[Optional[str]] = {card.artist, None}

    for name, number in names_cnums:
        for set_ in sets:
            for mvid in mvids:
                for artist in artists:
                    yield (set_, name, number, mvid, artist)


class ScryfallDataIndex:
    """Card and set indexes for scryfall data."""

    def __init__(self) -> None:
        self.id_to_card: Dict[UUID, ScryCard] = {}
        self.name_to_cards: Dict[str, List[ScryCard]] = {}
        self.setcode_to_cards: Dict[str, List[ScryCard]] = {}
        self.id_to_setindex: Dict[UUID, int] = {}
        self.setcode_to_set: Dict[str, ScrySet] = {}
        # snnma = Set, Name, (Collector) Number, Multiverse ID, Artist
        # TODO: convert to intersecting bitmap indexes
        # TODO: do we really need artist?
        self.snnma_to_id: Dict[
            Tuple[Optional[str], str, Optional[str], Optional[int], Optional[str]],
            Set[UUID],
        ] = {}

    def load_data(self, scrydata: ScryfallDataSet) -> None:
        """Load all cards and sets from a Scryfall data set."""
        self.id_to_card = {}
        self.id_to_setindex = {}
        self.setcode_to_set = {}

        self.snnma_to_id = collections.defaultdict(set)

        name_to_unsorted_cards: Dict[str, List[ScryCard]] = collections.defaultdict(
            list
        )
        setcode_to_unsorted_cards: Dict[str, List[ScryCard]] = {}

        for set_ in scrydata.sets:
            self.setcode_to_set[set_.code] = set_
            setcode_to_unsorted_cards[set_.code] = []

        for card in scrydata.cards:
            self.id_to_card[card.id] = card
            name_to_unsorted_cards[card.name].append(card)
            setcode_to_unsorted_cards[card.set].append(card)
            if not self.setcode_to_set[card.set].digital:
                for snnma in build_snnmas(card):
                    self.snnma_to_id[snnma].add(card.id)
        self.snnma_to_id = dict(self.snnma_to_id)

        for cards_list in name_to_unsorted_cards.values():
            cards_list.sort(key=name_card_sort_key)
        self.name_to_cards = dict(name_to_unsorted_cards)

        for cards_list in setcode_to_unsorted_cards.values():
            cards_list.sort(key=set_card_sort_key)
            self.id_to_setindex.update({c.id: i for i, c in enumerate(cards_list)})
        self.setcode_to_cards = dict(setcode_to_unsorted_cards)


class Oracle:
    """Container for an indexed Scryfall data set."""

    def __init__(self, scrydata: ScryfallDataSet) -> None:
        self._scrydata = scrydata
        self.cards = scrydata.cards
        self.sets = scrydata.sets
        self.index = ScryfallDataIndex()
        self.index.load_data(scrydata)



"""Interface definition for serializers."""

import abc
from pathlib import Path
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type

from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.indexes import Oracle


class Error(Exception):
    """Base error for serializers."""


class UnknownDialect(Exception):
    """Raised when an (extension, dialect) pair is requested."""


class DeserializationError(Error):
    """Raised when there is an error reading counts from a file."""


class SerializationDialect(metaclass=abc.ABCMeta):
    """Abstract interface for mtg ssm serialization dialect."""

    _EXT_DIALECT_DOC: ClassVar[Set[Tuple[str, str, str]]] = set()
    _EXT_DIALECT_TO_IMPL: ClassVar[
        Dict[Tuple[str, str], Type["SerializationDialect"]]
    ] = {}

    extension: ClassVar[Optional[str]] = None
    dialect: ClassVar[Optional[str]] = None

    def __init_subclass__(cls: Type["SerializationDialect"]) -> None:
        super().__init_subclass__()
        if cls.extension is not None and cls.dialect is not None:
            cls._EXT_DIALECT_DOC.add(
                (cls.extension, cls.dialect, cls.__doc__ or cls.__name__)
            )
            cls._EXT_DIALECT_TO_IMPL[(cls.extension, cls.dialect)] = cls

    @abc.abstractmethod
    def write(self, path: Path, collection: MagicCollection) -> None:
        """Write print counts to a file."""

    @abc.abstractmethod
    def read(self, path: Path, oracle: Oracle) -> MagicCollection:
        """Read print counts from file."""

    @classmethod
    def dialects(
        cls: Type["SerializationDialect"],
    ) -> List[Tuple[str, Optional[str], Optional[str]]]:
        """List of (extension, dialect, description) of registered dialects."""
        return sorted(
            (ext, dial or "", doc or "") for ext, dial, doc in cls._EXT_DIALECT_DOC
        )

    @classmethod
    def by_extension(
        cls: Type["SerializationDialect"],
        extension: str,
        dialect_mappings: Dict[str, str],
    ) -> Type["SerializationDialect"]:
        """Get a serializer class for a given extension and dialect mapping."""
        dialect = dialect_mappings.get(extension, extension)
        try:
            return cls._EXT_DIALECT_TO_IMPL[(extension, dialect)]
        except KeyError:
            raise UnknownDialect(
                f'File extension: "{extension}" dialect: "{dialect}" not found in registry'
            )



"""Scryfall data fetcher."""

from concurrent.futures import ProcessPoolExecutor
import gzip
import json
import os
import pickle
from typing import Any
from typing import List
from typing import Mapping
from typing import Union
from typing import cast
import uuid

import appdirs
import requests

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.scryfall import schema
from mtg_ssm.scryfall.models import ScryBulkData
from mtg_ssm.scryfall.models import ScryCard
from mtg_ssm.scryfall.models import ScryObject
from mtg_ssm.scryfall.models import ScryObjectList
from mtg_ssm.scryfall.models import ScrySet

DEBUG = os.getenv("DEBUG", "0")

APP_AUTHOR = "gwax"
APP_NAME = "mtg_ssm"
CACHE_DIR = appdirs.user_cache_dir(APP_NAME, APP_AUTHOR)

BULK_DATA_ENDPOINT = "https://api.scryfall.com/bulk-data"
SETS_ENDPOINT = "https://api.scryfall.com/sets"
BULK_TYPE = "default_cards"
OBJECT_CACHE_URL = "file://$CACHE/pickled_object"

CHUNK_SIZE = 8 * 1024 * 1024
DESERIALIZE_BATCH_SIZE = 50
_OBJECT_SCHEMA = schema.ScryfallUberSchema()

JSON = Union[str, int, float, bool, None, Mapping[str, Any], List[Any]]


def _cache_path(endpoint: str) -> str:
    cache_id = uuid.uuid5(uuid.NAMESPACE_URL, endpoint)
    return os.path.join(CACHE_DIR, str(cache_id))


def _fetch_endpoint(endpoint: str, *, dirty: bool, write_cache: bool = True) -> JSON:
    print(f"Retrieving {endpoint}")
    os.makedirs(CACHE_DIR, exist_ok=True)
    cache_path = _cache_path(endpoint)
    if not os.path.exists(cache_path):
        dirty = True
    if dirty:
        print(f"Fetching {endpoint}")
        response = requests.get(endpoint, stream=True)
        response.raise_for_status()
        if not write_cache:
            return response.json()
        print(f"Caching {endpoint}")
        with gzip.open(cache_path, "wb") as cache_file:
            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                cache_file.write(chunk)
    else:
        print("Reading cache")

    with gzip.open(cache_path, "rt", encoding="utf-8") as cache_file:
        return json.load(cache_file)


def _deserialize_object(obj_json: JSON) -> Union[ScryObject, List[ScryObject]]:
    return _OBJECT_SCHEMA.load(obj_json).data


def _deserialize_cards(card_jsons: List[JSON]) -> List[ScryCard]:
    cards_data: List[ScryCard]
    if DEBUG == "1":
        print("Process pool disabled")
        cards_data = []
        for card_json in card_jsons:
            try:
                cards_data.append(cast(ScryCard, _deserialize_object(card_json)))
            except Exception:
                print("Failed on: ", repr(card_json))
                raise
    else:
        with ProcessPoolExecutor() as executor:
            cards_futures = executor.map(
                _deserialize_object, card_jsons, chunksize=DESERIALIZE_BATCH_SIZE
            )
            cards_data = cast(List[ScryCard], list(cards_futures))
    return cards_data


def scryfetch() -> ScryfallDataSet:
    """Retrieve and deserialize Scryfall object data."""
    cached_bulk_json = None
    if os.path.exists(_cache_path(BULK_DATA_ENDPOINT)):
        cached_bulk_json = _fetch_endpoint(BULK_DATA_ENDPOINT, dirty=False)
    bulk_json = _fetch_endpoint(BULK_DATA_ENDPOINT, dirty=True, write_cache=False)
    cache_dirty = bulk_json != cached_bulk_json

    bulk_list: ScryObjectList = cast(ScryObjectList, _deserialize_object(bulk_json))
    sets_list = cast(
        ScryObjectList,
        _deserialize_object(_fetch_endpoint(SETS_ENDPOINT, dirty=cache_dirty)),
    )
    sets_data = cast(List[ScrySet], sets_list.data)
    while sets_list.has_more:
        sets_list = cast(
            ScryObjectList,
            _deserialize_object(
                _fetch_endpoint(str(sets_list.next_page), dirty=cache_dirty)
            ),
        )
        sets_data += cast(List[ScrySet], sets_list.data)

    bulk_data = cast(List[ScryBulkData], bulk_list.data)
    [cards_endpoint] = [bd.permalink_uri for bd in bulk_data if bd.type == BULK_TYPE]
    cards_json = cast(List[JSON], _fetch_endpoint(cards_endpoint, dirty=cache_dirty))

    _fetch_endpoint(BULK_DATA_ENDPOINT, dirty=cache_dirty, write_cache=True)

    if os.path.exists(_cache_path(OBJECT_CACHE_URL)) and not cache_dirty:
        try:
            with gzip.open(_cache_path(OBJECT_CACHE_URL), "rb") as object_cache:
                loaded_data = pickle.load(object_cache)
            if isinstance(loaded_data, ScryfallDataSet):
                return loaded_data
        except (OSError, pickle.UnpicklingError):
            pass
        print("Error reading object cache, falling back")

    print("Deserializing")
    cards_data = _deserialize_cards(cards_json)

    scryfall_data = ScryfallDataSet(sets=sets_data, cards=cards_data)
    with gzip.open(_cache_path(OBJECT_CACHE_URL), "wb") as object_cache:
        pickle.dump(scryfall_data, object_cache)
    return scryfall_data



"""Marshmallow schemas for Scryfall api data."""

import collections
from decimal import Decimal
from enum import EnumMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import marshmallow
from marshmallow import fields
from marshmallow_annotations import AnnotationSchema
from marshmallow_annotations import BaseConverter
from marshmallow_annotations import registry
from marshmallow_enum import EnumField
from marshmallow_oneofschema import OneOfSchema

from mtg_ssm.scryfall import models
from mtg_ssm.scryfall.third_party import marshmallow_fields

registry.register_field_for_type(models.URI, fields.Url)


@registry.field_factory(Decimal)
def decimal_converter(
    _converter: BaseConverter, _subtypes: Any, opts: Dict[str, Any]
) -> fields.Field:
    """Decimal field factory that always sets as_string."""
    return fields.Decimal(as_string=True, **opts)


@registry.field_factory(dict)
@registry.field_factory(Dict)
def dict_converter(
    converter: BaseConverter, subtypes: Any, opts: Dict[str, Any]
) -> fields.Field:
    """Dictionary field factory for marshmallow annotations."""
    sub_opts = opts.pop("_interior", ({}, {}))
    return marshmallow_fields.Dict(
        keys=converter.convert(subtypes[0], sub_opts[0]),
        values=converter.convert(subtypes[1], sub_opts[1]),
        **opts,
    )


class TupleField(fields.List):
    """Tuple field based on tuple() wrapping List field."""

    def _serialize(self, value: Any, attr: Any, obj: Any) -> Optional[Tuple[Any, ...]]:
        ret = super()._serialize(value, attr, obj)
        if ret is None:
            return None
        return tuple(ret)

    def _deserialize(
        self, value: Any, attr: Any, data: Any
    ) -> Optional[Tuple[Any, ...]]:
        ret = super()._deserialize(value, attr, data)
        if ret is None:
            return None
        return tuple(ret)


@registry.field_factory(Sequence)
@registry.field_factory(collections.abc.Sequence)
def sequence_converter(
    converter: BaseConverter, subtypes: Any, opts: Dict[str, Any]
) -> fields.Field:
    """Sequence field factory that treats sequences as lists."""
    sub_opts = opts.pop("_interior", {})
    return TupleField(converter.convert(subtypes[0], sub_opts), **opts)


def _register_enum(enum_class: EnumMeta) -> None:
    """Enum field registration helper function."""

    @registry.field_factory(enum_class)
    def _enum_converter(
        _converter: BaseConverter, _subtypes: Any, opts: Dict[str, Any]
    ) -> fields.Field:
        return EnumField(enum_class, by_value=True, **opts)


STR_ENUMS: Sequence[EnumMeta] = (
    models.ScryBorderColor,
    models.ScryCardFrame,
    models.ScryCardLayout,
    models.ScryRarity,
    models.ScryColor,
    models.ScryFormat,
    models.ScryFrameEffect,
    models.ScryGame,
    models.ScryLegality,
    models.ScrySetType,
)
for _str_enum in STR_ENUMS:
    _register_enum(_str_enum)


class BaseSchema(AnnotationSchema):
    """Shared marshmallow schema helper class."""

    @marshmallow.validates_schema(pass_original=True)
    def disallow_unknown_fields(
        self, _: Any, original_data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> None:
        """Check that no extra fields have been passed in."""
        if isinstance(original_data, (list, tuple)):
            unknowns = [sorted(set(o) - set(self.fields)) for o in original_data]
            if any(unknowns):
                raise marshmallow.ValidationError(
                    f"{type(self).__name__} Unknown fields: {repr(unknowns)}"
                )
        else:
            unknown = sorted(set(original_data) - set(self.fields))
            if unknown:
                raise marshmallow.ValidationError(
                    f"{type(self).__name__} Unknown fields: {repr(unknown)}"
                )

    @marshmallow.post_load
    def make_target(self, data: Dict[str, Any]) -> Any:
        """Post load, deserialize to target class."""
        return self.opts.target(**data)

    @marshmallow.post_dump
    def remove_none(  # pylint: disable=no-self-use
        self, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Post dump, strip null values."""
        return {k: v for k, v in data.items() if v is not None}


class PreviewBlockSchema(BaseSchema):
    """Schema for CardPreviewBlock."""

    class Meta:
        """Marshmallow configuration options."""

        target = models.CardPreviewBlock
        register_as_scheme = True
        strict = True


@registry.field_factory(models.ScryObject)
def _scryfall_converter(
    _converter: BaseConverter, _subtypes: Any, opts: Dict[str, Any]
) -> fields.Field:
    """Treat ScryfallObject references as OneOfSchema lookups."""
    return fields.Nested("ScryfallUberSchema", **opts)


MODELS: Sequence[Type[models.ScryObject]] = (
    models.ScryRelatedCard,
    models.ScryCardFace,
    models.ScryCard,
    models.ScrySet,
    models.ScryBulkData,
    models.ScryObjectList,
)
OBJECT_SCHEMAS: Dict[str, BaseSchema] = {}
for _model in MODELS:
    _meta = type(
        "Meta", (), {"target": _model, "register_as_scheme": True, "strict": True}
    )
    _schema_class = type(
        f"{_model.__name__}Schema",
        (BaseSchema,),
        {"Meta": _meta, "object": fields.String(dump_only=True)},
    )
    globals()[_schema_class.__name__] = _schema_class
    OBJECT_SCHEMAS[_model.object] = _schema_class()


class ScryfallUberSchema(OneOfSchema):
    """Lookup master schema for ScryfallObject subclasses."""

    class Meta:  # TODO: do we need this?
        """Marshmallow configuration options."""

        strict = True

    type_field = "object"
    type_schemas = OBJECT_SCHEMAS

    def get_obj_type(self, obj: models.ScryObject) -> str:
        return getattr(obj, self.type_field)



"""Legacy record lookup capabilities for older file versions."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from uuid import UUID

from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.mtg import util


class Error(Exception):
    """Base exception for this module."""


class NoMatchError(Error):
    """Raised when a matchor(Error):ing card cannot be found."""


class MultipleMatchError(Error):
    """Raised when multiple matching cards are found."""


COUNT_TYPE_TO_OLD_COUNT_TYPES: Dict[str, Set[str]] = {
    "nonfoil": {"nonfoil", "copies"},
    "foil": {"foil", "foils"},
}


def extract_counts(card_row: Dict[str, Any]) -> Dict[str, int]:
    """Convert old count type names to new count type names."""
    renamed_counts = {
        k: sum(int(card_row.get(t) or 0) for t in v)
        for k, v in COUNT_TYPE_TO_OLD_COUNT_TYPES.items()
    }
    return {k: v for k, v in renamed_counts.items() if v}


OTHER_SET_CODE_TO_SET_CODE = {
    # TODO: MOAR!
    "pMEI": ["phpr", "pdrc", "pbok", "pdtp", "pres", "purl", "pdp10", "pfrf", "pori"],
    "pPRE": ["psom", "pdgm", "pths", "pm15", "pktk", "pfrf", "pdtk", "pori", "pbfz"],
    "pJGP": [
        "jgp",
        "g99",
        "g00",
        "g01",
        "g02",
        "g03",
        "g04",
        "g05",
        "g06",
        "g07",
        "g08",
        "g09",
        "g11",
    ],
    "PO2": ["p02"],
    "pFNM": [
        "fnm",
        "f01",
        "f02",
        "f03",
        "f04",
        "f05",
        "f06",
        "f07",
        "f08",
        "f09",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
    ],
    "NMS": ["nem"],
    "pARL": ["pal99", "pal00", "pal01", "pal02", "pal03", "pal04", "pal05", "pal06"],
    "VAN": ["pvan"],
    "pGTW": ["pg07", "pg08"],
    "pMGD": ["p10e", "pnph", "pori"],
    "pLPA": ["pmbs", "pavr", "pktk"],
    "pWPN": ["pwwk", "pwp10", "proe"],
    "HOP": ["ohop", "phop"],
    "PC2": ["opc2", "phop"],
    "pWCQ": ["ppro"],
    "DD3_JVC": ["jvc"],
    "DD3_GVL": ["gvl"],
    "DD3_EVG": ["evg"],
    "DD3_DVD": ["dvd"],
    "FRF_UGIN": ["ugin"],
    "PCA": ["opca"],
}

PSUDONYM_TO_ARTIST: Dict[Optional[str], str] = {
    "William Murai": "Willian Murai",
    "Dave Seeley": "David Seeley",
}


def find_scryfall_id(card_row: Dict[str, str], oracle: Oracle) -> UUID:
    """Heuristically determine the scryfall id for a given input row."""
    set_code = card_row.get("set", "")
    set_codes = [set_code, set_code.lower()] + OTHER_SET_CODE_TO_SET_CODE.get(
        set_code, []
    )
    name = card_row.get("name", "")
    collector_number = card_row.get("number") or None
    mvid = int(card_row.get("multiverseid") or -1)
    artist = card_row.get("artist") or None
    artist = PSUDONYM_TO_ARTIST.get(artist, artist)
    print(
        f"Searching => Set: {set_code}; Name: {name}; Number: {collector_number}; MVID: {mvid}"
    )
    snnma_keys: List[
        Tuple[Optional[str], str, Optional[str], Optional[int], Optional[str]]
    ] = []
    for set_ in set_codes:
        snnma_keys += [
            (set_, name, collector_number, None, None),
            (set_, name, None, mvid, None),
            (set_, name, None, None, artist),
            (set_, name, None, None, None),
            (None, name, None, None, artist),
        ]
    seen = False
    for snnma_key in snnma_keys:
        found = oracle.index.snnma_to_id.get(snnma_key)
        if found:
            seen = True
            scryfall_id = None
            if len(found) == 1:
                [scryfall_id] = found
            if util.is_strict_basic(name):
                scryfall_id = sorted(found)[0]
            if scryfall_id is not None:
                found_card = oracle.index.id_to_card[scryfall_id]
                print(
                    f"Found ==> Set: {found_card.set}; Name: {found_card.name}; Number: {found_card.collector_number}; MVIDs: {found_card.multiverse_ids}"
                )
                return scryfall_id
    if seen:
        raise MultipleMatchError(f"Could not find scryfall card for row: {card_row}")
    raise NoMatchError(f"Could not find scryfall card for row: {card_row}")


def coerce_row(card_row: Dict[str, Any], oracle: Oracle) -> Dict[str, Any]:
    """Coerce an unknown older row into a current row.

    If this cannot be done precisely, a warning will be displayed.
    """
    coerced_counts = extract_counts(card_row)
    if not coerced_counts:
        return {}
    coerced_scryfall_id = find_scryfall_id(card_row, oracle)
    return {"scryfall_id": coerced_scryfall_id, **coerced_counts}



"""Data bundle definitions."""

from typing import List
from typing import NamedTuple
from typing import Set

from mtg_ssm.scryfall.models import ScryCard
from mtg_ssm.scryfall.models import ScrySet
from mtg_ssm.scryfall.models import ScrySetType


class ScryfallDataSet(NamedTuple):
    """Bundle for storing Scryfall data."""

    sets: List[ScrySet]
    cards: List[ScryCard]


def remove_digital(scryfall_data: ScryfallDataSet) -> ScryfallDataSet:
    """Filter a ScryfallDataSet to remove all digital only sets and cards."""
    rejected_setcodes = set()
    accepted_sets = []
    for set_ in scryfall_data.sets:
        if set_.digital:
            rejected_setcodes.add(set_.code)
        else:
            accepted_sets.append(set_)
    accepted_cards = [c for c in scryfall_data.cards if c.set not in rejected_setcodes]
    return ScryfallDataSet(sets=accepted_sets, cards=accepted_cards)


def filter_set_types(
    scryfall_data: ScryfallDataSet, set_types: Set[ScrySetType]
) -> ScryfallDataSet:
    """Filter a ScryfallDataSet to include only specified set types."""
    rejected_setcodes = set()
    accepted_sets = []
    for set_ in scryfall_data.sets:
        if set_.set_type in set_types:
            accepted_sets.append(set_)
        else:
            rejected_setcodes.add(set_.code)
    accepted_cards = [c for c in scryfall_data.cards if c.set not in rejected_setcodes]
    return ScryfallDataSet(sets=accepted_sets, cards=accepted_cards)



#!/usr/bin/env python3
"""Script for managing magic card spreadsheets."""

import argparse
import datetime as dt
from pathlib import Path
import tempfile
from typing import Dict
from typing import List
from typing import Set

import mtg_ssm
from mtg_ssm.containers import bundles
from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.scryfall import fetcher
from mtg_ssm.scryfall.models import ScrySetType
import mtg_ssm.serialization.interface as ser_interface


def epilog() -> str:
    """Generate the argparse help epilog with dialect descriptions."""
    dialect_docs = "available dialects:\n"
    ext_dia_desc = ser_interface.SerializationDialect.dialects()
    for extension, dialect, description in ext_dia_desc:
        dialect_docs += f"  {extension:<8} {dialect:<12} {description}\n"
    return dialect_docs


def set_type_list(value: str) -> Set[ScrySetType]:
    """Argparse type to convert a string to a set of Scryfall Set Types."""
    set_types = set()
    for set_str in value.split(","):
        try:
            set_types.add(ScrySetType(set_str))
        except ValueError as err:
            msg = (
                f"{set_str} in {value} is not a valid set_type, please use a commas separated list of values from: "
                + ", ".join(ScrySetType)
            )
            raise argparse.ArgumentTypeError(msg) from err
    return set_types


def get_args(args: List[str] = None) -> argparse.Namespace:
    """Parse and return application arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Magic collection Spreadsheet Manager",
        epilog=epilog(),
    )
    parser.add_argument("--version", action="version", version=mtg_ssm.__version__)

    parser.add_argument(
        "--include-digital",
        default=False,
        action="store_true",
        help="Include digital only sets (e.g. Vintage Masters)",
    )

    parser.add_argument(
        "-d",
        "--dialect",
        nargs=2,
        metavar=("EXTENSION", "DIALECT"),
        action="append",
        default=[],
        help="Mapping of file extensions to serializer dialects. "
        "May be repeated for multiple different extensions.",
    )

    default_set_types = set(ScrySetType) - {ScrySetType.MEMORABILIA, ScrySetType.TOKEN}
    parser.add_argument(
        "--set-types",
        default=",".join(default_set_types),
        type=set_type_list,
        help="List of set types to include as a comma separted list of values from: "
        + ", ".join(ScrySetType),
    )

    # Commands
    subparsers = parser.add_subparsers(dest="action", title="actions")
    subparsers.required = True

    create = subparsers.add_parser(
        "create", aliases=["c"], help="Create a new, empty collection spreadsheet"
    )
    create.set_defaults(func=create_cmd)
    create.add_argument("collection", type=Path, help="Filename for the new collection")

    update = subparsers.add_parser(
        "update",
        aliases=["u"],
        help="Update cards in a collection spreadsheet, preserving counts",
    )
    update.set_defaults(func=update_cmd)
    update.add_argument(
        "collection", type=Path, help="Filename for the collection to update"
    )

    merge = subparsers.add_parser(
        "merge",
        aliases=["m"],
        help="Merge one or more collection spreadsheets into another. May also be used for format conversions.",
    )
    merge.set_defaults(func=merge_cmd)
    merge.add_argument(
        "collection", type=Path, help="Filename for the target collection"
    )
    merge.add_argument(
        "imports",
        nargs="+",
        type=Path,
        help="Filename(s) for collection(s) to import/merge counts from",
    )

    diff = subparsers.add_parser(
        "diff",
        aliases=["d"],
        help="Create a collection from the differences between two other collections",
    )
    diff.set_defaults(func=diff_cmd)
    diff.add_argument(
        "left",
        type=Path,
        help="Filename for first collection to diff (positive counts)",
    )
    diff.add_argument(
        "right",
        type=Path,
        help="Filename for second collection to diff (negative counts)",
    )
    diff.add_argument(
        "output", type=Path, help="Filename for result collection of diff"
    )

    parsed_args = parser.parse_args(args=args)
    parsed_args.dialect = dict(parsed_args.dialect)
    return parsed_args


def get_oracle(set_types: Set[ScrySetType], include_digital: bool) -> Oracle:
    """Get a card_db with current mtgjson data."""
    scrydata = fetcher.scryfetch()
    scrydata = bundles.filter_set_types(scrydata, set_types)
    if not include_digital:
        scrydata = bundles.remove_digital(scrydata)
    return Oracle(scrydata)


def get_serializer(
    dialect_mapping: Dict[str, str], path: Path
) -> ser_interface.SerializationDialect:
    """Retrieve a serializer compatible with a given filename."""
    extension = path.suffix.lstrip(".")
    serialization_class = ser_interface.SerializationDialect.by_extension(
        extension, dialect_mapping
    )
    return serialization_class()


def get_backup_path(path: Path) -> Path:
    """Given a filename, return a timestamped backup name for the file."""
    now = dt.datetime.now()
    return path.parent / f"{path.stem}.{now:%Y%m%d_%H%M%S}{path.suffix}"


def write_file(
    serializer: ser_interface.SerializationDialect,
    collection: MagicCollection,
    path: Path,
) -> None:
    """Write print counts to a file, backing up existing target files."""
    if not path.exists():
        print(f"Writing collection to file: {path}")
        serializer.write(path, collection)
    else:
        backup_path = get_backup_path(path)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir) / path.name
            print("Writing to temporary file.")
            serializer.write(temp_path, collection)
            print(f"Backing up existing file to: {backup_path}")
            path.replace(backup_path)
            print(f"Writing collection: {path}")
            temp_path.replace(path)


def create_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Create a new, empty collection."""
    collection = MagicCollection(oracle=oracle, counts={})
    serializer = get_serializer(args.dialect, args.collection)
    write_file(serializer, collection, args.collection)


def update_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Update an existing collection, preserving counts."""
    serializer = get_serializer(args.dialect, args.collection)
    print(f"Reading counts from {args.collection}")
    collection = serializer.read(args.collection, oracle)
    write_file(serializer, collection, args.collection)


def merge_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Merge counts from one or more inputs into a new/existing collection."""
    coll_serializer = get_serializer(args.dialect, args.collection)
    collection = MagicCollection(oracle=oracle, counts={})
    if args.collection.exists():
        print(f"Reading counts from {args.collection}")
        collection = coll_serializer.read(args.collection, oracle)
    for import_path in args.imports:
        input_serializer = get_serializer(args.dialect, import_path)
        print(f"Merging counts from {import_path}")
        collection += input_serializer.read(import_path, oracle)
    write_file(coll_serializer, collection, args.collection)


def diff_cmd(args: argparse.Namespace, oracle: Oracle) -> None:
    """Diff two collections, putting the output in a third."""
    left_serializer = get_serializer(args.dialect, args.left)
    right_serializer = get_serializer(args.dialect, args.right)
    output_serializer = get_serializer(args.dialect, args.output)
    print(f"Diffing counts between {args.left} and {args.right}")
    diff_collection = left_serializer.read(args.left, oracle) - right_serializer.read(
        args.right, oracle
    )
    write_file(output_serializer, diff_collection, args.output)


def main() -> None:
    """Get args and run the appropriate command."""
    args = get_args()
    oracle = get_oracle(args.set_types, args.include_digital)
    args.func(args, oracle)


if __name__ == "__main__":
    main()



"""Helpers for tracking collection counts."""

import collections
import enum
from typing import Any
from typing import Dict
from typing import Iterable
from typing import MutableMapping
from uuid import UUID

from mtg_ssm.containers import legacy
from mtg_ssm.containers.indexes import Oracle


class CountType(enum.Enum):
    """Enum for possible card printing types (nonfoil, foil)."""

    nonfoil = enum.auto()
    foil = enum.auto()


ScryfallCardCount = Dict[UUID, MutableMapping[CountType, int]]
"""Mapping from scryfall id to card printing type to count."""


def aggregate_card_counts(
    card_rows: Iterable[Dict[str, Any]], oracle: Oracle
) -> ScryfallCardCount:
    """Extract card counts from card rows."""
    card_counts: ScryfallCardCount = {}
    for card_row in card_rows:
        if "scryfall_id" not in card_row:
            card_row = legacy.coerce_row(card_row, oracle)
        if not card_row:
            continue
        scryfall_id = card_row["scryfall_id"]
        if not isinstance(scryfall_id, UUID):
            scryfall_id = UUID(scryfall_id)
        counts = card_counts.get(scryfall_id, {})
        for count_type in CountType:
            value = int(card_row.get(count_type.name) or 0)
            if value:
                counts[count_type] = value + counts.get(count_type, 0)
        if counts:
            card_counts[scryfall_id] = counts
    return card_counts


def merge_card_counts(*card_counts_args: ScryfallCardCount) -> ScryfallCardCount:
    """Merge two sets of card_counts."""
    merged_counts: ScryfallCardCount = collections.defaultdict(collections.Counter)
    for card_counts in card_counts_args:
        for card_id, counts in card_counts.items():
            merged_counts[card_id].update(counts)
    return dict(merged_counts)


def diff_card_counts(
    left: ScryfallCardCount, right: ScryfallCardCount
) -> ScryfallCardCount:
    """Subtract right print counts from left print counts."""
    diffed_counts: ScryfallCardCount = collections.defaultdict(dict)
    for card_id in left.keys() | right.keys():
        left_counts = left.get(card_id, {})
        right_counts = right.get(card_id, {})
        card_counts = {
            k: left_counts.get(k, 0) - right_counts.get(k, 0)
            for k in left_counts.keys() | right_counts.keys()
        }
        card_counts = {k: v for k, v in card_counts.items() if v}
        if card_counts:
            diffed_counts[card_id] = card_counts
    return diffed_counts
